import json

import numpy as np
import tensorflow as tf
from tensorflow.keras.optimizers.schedules import *


class LiveLrSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
	"""
	Updates learning rate schedule based on config file during the training process.
	"""
	def __init__(self, check_for_update_interval, lr_filename='current_lr.conf', custom_objects={}):
		"""
		:param check_for_update_interval: The interval the lr will be updated. Measured in **steps**.
		:param lr_filename: The file to read config from.
		:param custom_objects:
			The dict passed to `deserialize` function while loading lr schedule.
			This is typically for custom children of tf.keras.optimizers.schedules.LearningRateSchedule.
		"""
		self.lr_filename = lr_filename
		self.check_for_update_interval = check_for_update_interval
		self.custom_objects = custom_objects

		self.lr_file = open(lr_filename, 'r')
		self.base_schedule = None
		self.logger = tf.get_logger()

	def __call__(self, step):
		return tf.numpy_function(self.call, [step], tf.float32)

	def load_config(self):
		"""
		Tries to load config file and update `base_schedule`. Logs failures with logger.error.
		"""
		try:
			self.lr_file.seek(0)
			config = json.loads(self.lr_file.read())
			self.base_schedule = deserialize(config, self.custom_objects)
		except Exception as err:
			self.logger.error('LiveLrSchedule error: %s', str(err), exc_info=True)
			self.logger.error('LiveLrSchedule keeps schedule unchanged')

	def call(self, step):
		if step % self.check_for_update_interval == 0:
			self.load_config()

		return np.float32(self.base_schedule(step))



import os
from unittest import TestCase
from tempfile import NamedTemporaryFile
from pathlib import Path

import numpy as np
from tensorflow.keras import Sequential
from tensorflow.keras.optimizers import SGD
from tensorflow.keras.optimizers.schedules import LearningRateSchedule
from tensorflow.keras.callbacks import LambdaCallback
import tensorflow.keras.layers as layers

from tf_livepatch_lr.livepatch_lr import LiveLrSchedule


class ConstantLrSchedule(LearningRateSchedule):
	def __init__(self, c):
		super(ConstantLrSchedule, self).__init__()
		self.c = c

	def __call__(self, step):
		return self.c

	def get_config(self):
		return {'c': self.c}


class TestLiveLrSchedule(TestCase):

	def test_call(self):
		initial_config_file = open(Path('livepatch_lr') / 'polynomial_decay.json')
		last_config_file = open(Path('livepatch_lr') / 'constant.json')

		temp_config_file = NamedTemporaryFile('w+', delete=False)
		temp_config_file.write(initial_config_file.read())
		temp_config_file.close()

		self.xs = np.ones((1, 1, 1), np.float32)
		self.ys = self.xs.copy()
		self.model = Sequential([
			layers.Dense(1)
		])

		def on_epoch_end(epoch, logs):
			if epoch == 5:
				with open(temp_config_file.name, 'w') as f:
					f.truncate(0)
					f.seek(0)
					f.write(last_config_file.read())

		change_config_callback = LambdaCallback(on_epoch_end=on_epoch_end)

		lr = LiveLrSchedule(8, temp_config_file.name, custom_objects={'ConstantLrSchedule': ConstantLrSchedule})
		self.model.compile(SGD(lr), 'mse', run_eagerly=False)
		self.model.fit(self.xs, self.ys, epochs=10, callbacks=[change_config_callback])
		initial_config_file.close()
		last_config_file.close()
		os.unlink(temp_config_file.name)

		self.assertEqual(lr.base_schedule.__class__.__name__, 'ConstantLrSchedule')



from setuptools import setup


install_requires = [
	'numpy',
	'tensorflow'
]

tests_require = [
	'pathlib'
]

setup(
	name = 'tf_livepatch_lr',
	version = '0.0.2',
	url = 'https://github.com/andrewerf/tf_livepatch_lr',
	author = 'Andrew Aralov',
	author_email = 'andrew-aralov@yandex.ru',
	packages = ['tf_livepatch_lr', 'tf_livepatch_lr.test'],
	license = 'LICENSE.txt',
	description = 'Simple project which allows you to change the learning rate schedule '
				  'without stopping the training process (tensorflow only)',
	long_description_content_type = 'text/markdown',
	long_description = open('README.md').read(),
	install_requires = install_requires,
	tests_require = tests_require,

	classifiers = [
		'Programming Language :: Python :: 3 :: Only',
		'License :: OSI Approved :: MIT License',
		'Intended Audience :: Developers'
	]
)
